import requests
import ipaddress
from cnaas_sync.utils import get_number_of_pages

# from utils import get_number_of_pages


class NAV:
    def __init__(self, base_url, api_token, verbose=True):
        self.base_url = base_url
        self.api_token = api_token
        self.headers = {"Authorization": "Token " + self.api_token}
        self.verbose = verbose

    def _print(self, value):
        if self.verbose:
            print(value)

    def get_all_locations(self):
        """Fetch all locations from the API with proper error handling."""
        url = f"{self.base_url}/api/location/"
        try:
            response = requests.get(url, headers=self.headers, timeout=10)
            response.raise_for_status()  # Raise an HTTPError for bad responses (4xx and 5xx)
            return response.json()
        except requests.exceptions.RequestException as e:
            return {"error": str(e), "status": "failed"}

    def add_location(self, name):
        """Add a location if it does not already exist."""
        url = f"{self.base_url}/api/location/"
        data = {"id": name, "description": f"Location '{name}' created by sync script"}

        try:
            # Fetch existing locations
            locations_response = self.get_all_locations()
            if "results" not in locations_response:
                raise ValueError(
                    "Unexpected response structure from get_all_locations()"
                )

            # Check if location already exists
            if any(room.get("id") == name for room in locations_response["results"]):
                return name  # Location already exists

            # Create the new location
            response = requests.post(url, headers=self.headers, json=data, timeout=10)
            response.raise_for_status()

            return response.json().get("id")
        except requests.exceptions.RequestException as e:
            if hasattr(self, "logger"):
                self.logger.error(f"Error creating location '{name}': {e}")
            return None
        except ValueError as e:
            if hasattr(self, "logger"):
                self.logger.error(str(e))
            return None

    def add_room(self, room: str, location: str):
        """Add a room if it does not already exist."""
        url = f"{self.base_url}/api/room/"
        data = {
            "id": room,
            "description": f"Room '{room}' created at '{location}' by sync script",
            "location": location,
        }
        try:
            # Fetch existing rooms
            rooms_response = self.get_all_rooms()
            if not isinstance(rooms_response, list):
                raise ValueError("Unexpected response structure from get_all_rooms()")

            # Check if room already exists
            if any(room_.get("id") == room for room_ in rooms_response):
                return room

            # Create the new room
            response = requests.post(url, headers=self.headers, json=data, timeout=10)
            response.raise_for_status()  # Raise error for 4xx/5xx responses
            return response.json()
        except Exception as e:
            print(f"Error creating room '{room}' at '{location}': {e}")
            return None

    def switch_name(self, s: str) -> str:
        category_map = {
            "CORE": "GSW",
            "DIST": "GSW",
            "ACCESS": "SW",
        }
        s_lower = s.lower()
        for keyword, category in category_map.items():
            if keyword.lower() in s_lower:
                return category
        return None

    def cidr_to_ip(self, cidr: str) -> str:
        # Create an IPv4 network object from the CIDR notation
        network = ipaddress.IPv4Network(cidr, strict=False)

        # Return the network's IP address
        return str(network.network_address)

    def add_new_device(
        self,
        management_ip,
        roomid,
        organizationid,
        categoryid,
        hostname,
        device_type,
        nav_devices,
    ):
        data = {
            "ip": self.cidr_to_ip(management_ip.address) if management_ip else None,
            "roomid": roomid,
            "organizationid": organizationid,
            "categoryid": self.switch_name(categoryid),
            "sysname": hostname,
            "profiles": [1],
            "data": {"DEVICE_TYPE": device_type},
        }

        url = f"{self.base_url}/api/netbox/"

        if any(device["ip"] == management_ip for device in nav_devices):
            self._print(f"Switch {hostname} found in NAV!")
            return None

        self._print(f"Adding {hostname}")
        response = requests.post(url, headers=self.headers, json=data)

        if response.status_code != 201:
            self._print(
                f"Could not create device! Status: {response.status_code}, Response: {response.text}"
            )
            return False

        return response.json()

    def get_all_rooms(self):
        """Fetch all rooms from the API with proper error handling."""
        url = f"{self.base_url}/api/room/"
        try:
            response = requests.get(url, headers=self.headers, timeout=10)
            response.raise_for_status()
            rooms = response.json().get("results", [])
            for room in rooms:
                self._print(f"Room {room.get('id', 'UNKNOWN')} found in NAV!")
            return rooms
        except requests.exceptions.RequestException as e:
            print(f"Failed to fetch rooms from NAV: {e}")
            return []

    def get_all_devices(self, page_size=100):
        res = requests.get(
            self.base_url + f"/api/1/netbox/?page=1&page_size={page_size}",
            headers=self.headers,
        )
        if not res.status_code == 200:
            print(f"Failed to fetch devices from NAV: {res.status_code}")
            return False
        data_json = res.json()
        total_count = int(data_json["count"])
        num_pages = get_number_of_pages(total_count, page_size)
        accumulated_res = data_json["results"]
        for page in range(2, num_pages + 1):
            res_tmp = requests.get(
                self.base_url + f"/api/1/netbox/?page={page}&page_size={page_size}",
                headers=self.headers,
            )
            if "results" not in res_tmp.json():
                print(
                    "Failed to fetch devices from NMS! Status Code: "
                    + str(res_tmp.status_code)
                )
                continue
            accumulated_res.extend(res_tmp.json()["results"])  # extend vs +=
        return accumulated_res

    def add_new_managed_device(
        self, hostname, switches, category_id, management_ip, room, device_type
    ):
        data = {
            "ip": management_ip[0],
            "roomid": room,
            "organizationid": "SUNET",
            "categoryid": category_id,
            "sysname": f"{hostname}",
            "profiles": [1],
            "data": {"DEVICE_TYPE": device_type},
        }
        for device in switches:
            if device["ip"] == management_ip[0]:
                self._print(f"Switch {device['sysname']} found in NAV!")
                return False
        self._print(f"Adding {hostname}")
        res2 = requests.post(
            self.base_url + "/api/1/netbox/",
            headers=self.headers,
            json=data,
        )
        if res2.status_code != 201:
            print(f"Could not create device!  {res2}")
            return False

    def update(self, ni_devices, nms):
        all_nav_devices = self.get_all_devices()
        self._print("**" * 20)
        for device in ni_devices:
            if device["node"]["operational_state"] != "In service":
                continue
            hostname = device["node_name"]
            device_type = nms.get_device_type(hostname)
            if device_type == "CORE" or device_type == "DIST":
                category_id = "GSW"
            elif device_type == "ACCESS":
                category_id = "SW"
            else:
                continue
            management_ip = device["node"]["ip_addresses"]
            self.add_new_managed_device(
                hostname,
                all_nav_devices,
                category_id,
                management_ip,
                "myroom",
                device_type,
            )
            self._print("**" * 20)
