import os
import pynetbox
import traceback
from typing import Optional
from slugify import slugify
from cnaas_sync.utils import remove_prefix_suffix, print_section, print_color
from cnaas_sync.schema import (
    InterfaceModel,
    DeviceModel,
    DeviceTypeModel,
    IpAddressModel,
    DeviceRole,
    DeviceStatusChoices,
    IPAddressStatusChoices,
    DeviceName,
)


class NetBoxAPI:
    def __init__(self, netbox_url: str, api_token: str):
        # Initialize the pynetbox API instance
        self.nb = pynetbox.api(netbox_url, token=api_token)

    def get_tenant(self, tenant_name: str):
        """Fetch tenant ID by name, raise error if tenant not found."""
        tenant = self.nb.tenancy.tenants.get(name=tenant_name)
        if not tenant:
            raise ValueError(f"Tenant '{tenant_name}' not found in NetBox.")
        return tenant

    def get_site(self, tenant_id: str):
        """Fetch the site that belongs to the tenant, create a new site "<tenant>-default" if none exists."""
        sites = self.nb.dcim.sites.filter(tenant_id=tenant_id)
        if sites:
            return list(sites)[0]
        return None

    def get_or_create_site(self, tenant_id: str):
        """Fetch the site that belongs to the tenant, create a new site "<tenant>-default" if none exists."""
        sites = self.nb.dcim.sites.filter(tenant_id=tenant_id)
        if sites:
            return list(sites)[0]

        tenant = self.nb.tenancy.tenants.get(id=tenant_id)
        new_site_name = f"{str(tenant.name).lower()}-default"
        new_site = self.nb.dcim.sites.create({"name": new_site_name, "slug": slugify(new_site_name), "tenant": tenant.id})
        return new_site

    def get_devices_by_tenant(self, tenant_id: str):
        """Fetch all devices associated with the tenant."""
        devices = self.nb.dcim.devices.filter(tenant_id=tenant_id)
        return list(devices)

    def get_devices(self, tenant_id: str, status: str, manufacturer: str):
        """Fetch all active Arista devices associated with the tenant."""
        devices = self.nb.dcim.devices.filter(
            tenant_id=tenant_id, status=status, manufacturer=manufacturer
        )
        return list(devices)

    def get_device_roles(self):
        """Fetch all device types and return them in a list."""
        device_roles = self.nb.dcim.device_roles.all()
        return list(device_roles)

    def get_ipam_prefix_by_parent_tenant(self, tenant_id: str, management_ip: str):
        """Fetch IPAM prefixes associated with the tenant."""
        prefixes = self.nb.ipam.prefixes.filter(
            tenant_id=tenant_id, contains=management_ip
        )
        return list(prefixes)

    def get_device_types(self):
        """Fetch all device types and return them in a list."""
        device_types = self.nb.dcim.device_types.all()
        return list(device_types)

    def get_device_types_by_model(self, model: str, manufacturer: str):
        """Fetch all Arista 7280SR-48C6 device types."""
        device_types = self.nb.dcim.device_types.filter(
            model=model, manufacturer=manufacturer
        )
        return list(device_types)

    def get_test_customs(self):
        """Fetch all device types and return them in a list."""
        return list(self.nb.extras.config_contexts.all())

    def create_device_type(
        self, device_type: DeviceTypeModel, dry_run: bool = False, verbose: bool = False
    ):
        """Create and return a device type."""
        new_device = None
        if dry_run:
            print(f"Dry run: Creating device type {device_type.model}")
        else:
            new_device = self.nb.dcim.device_types.create(device_type.model_dump())
            print(f"Created new device type: {new_device}") if verbose else None
        return new_device

    def create_device(self, device: DeviceModel, dry_run: bool = False):
        """Create a new device in NetBox."""
        try:
            if dry_run:
                print(
                    f"Dry run: Creating device {device.name} with type {device.device_type} for tenant {device.tenant} on site {device.site}"
                )
                return device
            new_device = self.nb.dcim.devices.create(device.model_dump())
            return new_device
        except Exception as e:
            print(f"Error creating device: {e}")
            return None

    def create_interface(self, interface: InterfaceModel, dry_run: bool = False):
        """Create a new interface on a device in NetBox."""
        try:
            if dry_run:
                print(
                    f"Dry run: Creating interface {interface.name} on device {interface.device}"
                )
                return interface
            device = self.nb.dcim.devices.get(name=interface.device)
            if not device:
                raise ValueError(f"Device {interface.device} not found")
            interface_dict = interface.model_dump(exclude={"id"})
            new_interface = self.nb.dcim.interfaces.create(interface_dict)
            return new_interface
        except Exception as e:
            print(f"Error creating interface: {e}")
            return None

    def create_ip_address(self, ip_address: IpAddressModel, dry_run: bool = False):
        """Create a new IP address in NetBox."""
        try:
            if dry_run:
                print(
                    f"Dry run: Creating IP address {ip_address.address} for tenant {ip_address.tenant} on interface.id {ip_address.assigned_object_id}"
                )
                return ip_address
            vrf = self.nb.ipam.vrfs.get(name=ip_address.vrf)
            if not vrf:
                raise ValueError(f"VRF {ip_address.vrf} not found")

            tenant = self.nb.tenancy.tenants.get(name=ip_address.tenant)
            if not tenant:
                raise ValueError(f"Tenant {ip_address.tenant} not found")

            ip_address = self.nb.ipam.ip_addresses.create(ip_address.model_dump())
            return ip_address
        except Exception as e:
            print(f"Error creating IP address: {e}")
            return None

    def get_interface_by_name(
        self, interface_name: str, device_name: Optional[str] = None
    ):
        """Fetch interface by name and optionally filter by device name."""
        try:
            query_params = {"name": interface_name}
            if device_name:
                query_params["device"] = device_name
            return self.nb.dcim.interfaces.get(**query_params)
        except Exception as e:
            print(f"Error retrieving interface: {e}")
            return None

    def get_vrf_by_tenant(self, tenant):
        """Fetch VRF by tenant."""
        try:
            vrf = self.nb.ipam.vrfs.filter(tenant_id=tenant.id, name__ic=tenant.name)
            vrf_list = list(vrf)
            return vrf_list[0] if len(vrf_list) > 0 else None
        except Exception as e:
            print(f"Error fetching VRF with tenant {tenant}: {e}")
            return None

    def create_device_types_if_none_exist(
        self,
        netbox_device_types,
        nms_devices,
        dry_run: bool = False,
        verbose: bool = False,
    ):
        # Use to keep track of (model, vendor) pair
        seen_pairs = set()

        netbox_device_models = set(
            d.part_number for d in netbox_device_types if d.part_number
        )
        nms_devices_models = set(d["model"] for d in nms_devices)
        type_name_not_in_netbox = nms_devices_models.difference(netbox_device_models)

        # Filter to ensure no two elements have the same (model, vendor) pair
        types_not_in_netbox = filter(
            lambda x: x["model"] in type_name_not_in_netbox
            and not (
                (x["model"], x["vendor"]) in seen_pairs
                or seen_pairs.add((x["model"], x["vendor"]))
            ),
            nms_devices,
        )
        types_not_in_netbox = list(types_not_in_netbox)

        if dry_run or verbose:
            synced_message = "ALL DEVICE TYPES ARE SYNCED IN NETBOX"
            unsynced_message = "DEVICE TYPES NOT FOUND IN NETBOX"

            # Select the appropriate message and color based on the presence of unsynced device types
            section_message = (
                unsynced_message if types_not_in_netbox else synced_message
            )
            section_color = "blue" if types_not_in_netbox else "green"
            print_section(section_message, color=section_color)
            print_color(
                set(d["model"] for d in types_not_in_netbox), color="red"
            ) if types_not_in_netbox else None

        # To create a new device type, we only need the (model, vendor) pairs from NMS
        for dev_type in types_not_in_netbox:
            try:
                # Extract and process model and vendor
                model = dev_type.get("model")
                vendor = dev_type.get("vendor")

                if not model or not vendor:
                    print_color(
                        f"Skipping device type due to missing model or vendor: {dev_type}",
                        color="yellow",
                    )
                    continue

                trimmed_model = remove_prefix_suffix(model)
                dtype = DeviceTypeModel.create(
                    manufacturer=vendor, model=trimmed_model, part_number=model
                )
                self.create_device_type(dtype, dry_run=dry_run)

                print_color(
                    f"Successfully created device type: {trimmed_model} (Vendor: {vendor})",
                    color="green",
                ) if verbose else None

            except Exception as e:
                print_color(
                    f"Failed to create device type for: {dev_type}. Error: {e}",
                    color="red",
                )

    def create_device_if_none_exist(
        self,
        netbox_devices,
        nms_devices,
        tenant,
        site,
        dry_run: bool = False,
        verbose: bool = False,
    ):
        netbox_device_names = set(d.name for d in netbox_devices)
        nms_device_names = set(d["hostname"] for d in nms_devices)

        device_names_not_in_netbox = nms_device_names.difference(netbox_device_names)
        devices_not_in_netbox = filter(
            lambda x: x["hostname"] in device_names_not_in_netbox, nms_devices
        )
        devices_not_in_netbox = list(devices_not_in_netbox)

        if dry_run or verbose:
            synced_message = "ALL DEVICE ARE SYNCED IN NETBOX"
            unsynced_message = "DEVICES NOT FOUND IN NETBOX"

            # Select the appropriate message and color based on the presence of unsynced device types
            section_message = (
                unsynced_message if devices_not_in_netbox else synced_message
            )
            section_color = "blue" if devices_not_in_netbox else "green"
            print_section(section_message, color=section_color)
            print_color(
                set(d["hostname"] for d in devices_not_in_netbox), color="red"
            ) if devices_not_in_netbox else None

        for dev in devices_not_in_netbox:
            created_devices = []
            created_interfaces = []
            created_ip_addresses = []
            try:
                new_device = {
                    "name": dev["hostname"],
                    "role": DeviceRole.from_string(dev["device_type"]),
                    "site": site.name,
                    "type": remove_prefix_suffix(dev["model"]),
                    "tenant": tenant.name,
                    "manufacturer": dev["vendor"],
                    "status": DeviceStatusChoices.STATUS_STAGED,
                    "serial": dev["serial"],
                }
                device_model = DeviceModel.create(**new_device)
                created_device = self.create_device(device_model, dry_run=dry_run)
                created_devices.append(created_device)

                # Create an interface
                interface_data = {
                    "device": DeviceName(name=dev["hostname"]),
                    "name": "MGMT",
                    "type": "virtual",
                }
                interface_model = InterfaceModel(**interface_data)
                created_interface = self.create_interface(
                    interface_model, dry_run=dry_run
                )
                created_interfaces.append(created_interface)

                # Create ip-address
                tenant_vrf = self.get_vrf_by_tenant(tenant=tenant)
                ip_addr = {
                    "address": dev["management_ip"],
                    "vrf": tenant_vrf.name,
                    "tenant": tenant.name,
                    "status": IPAddressStatusChoices.STATUS_ACTIVE,
                    "assigned_object_type": "dcim.interface",
                    "assigned_object_id": created_interface.id,
                }
                ip_addr_model = IpAddressModel.create(**ip_addr)
                created_ip_address = self.create_ip_address(
                    ip_addr_model, dry_run=dry_run
                )
                created_ip_addresses.append(created_ip_address)
            except Exception as e:
                # Rollback created devices, interfaces, and IPs if any operation fails
                print(f"Error occurred: {e}. Rolling back changes...")

                # Delete created interfaces and IPs in reverse order of creation
                for ip in created_ip_addresses:
                    if ip:
                        self.nb.ipam.ip_addresses.delete(ip.id)
                for interface in created_interfaces:
                    if interface:
                        self.nb.dcim.interfaces.delete(interface.id)
                for device in created_devices:
                    if device:
                        self.nb.dcim.devices.delete(device.id)

                # Reraise the exception after rollback
                # raise

    def test(self, tenant_name: str):
        """Test function to fetch and display tenant data."""
        try:
            tenant = self.get_tenant(tenant_name)
            site = self.get_or_create_site(tenant.id)
            devices = self.get_devices_by_tenant(tenant.id)
            device_types = self.get_device_types()

            print(f"Tenant ID for '{tenant_name}': {tenant.id}")
            print(f"Site for tenant: {site.name}")
            print(
                f"Devices for tenant '{tenant_name}': {[device.name for device in devices]}"
            )
            print(
                f"Device types: {[device_type.model for device_type in device_types]}"
            )

        except ValueError as e:
            print(e)

    def fetch_netbox_data(
        self, tenant_name, dry_run: bool = False, verbose: bool = False
    ):
        try:
            # Fetch the tenant
            tenant = self.get_tenant(tenant_name)
            if not tenant:
                raise ValueError(f"Tenant '{tenant_name}' not found in NetBox.")

            # Get or create the site for the tenant
            site = self.get_or_create_site(tenant.id)

            # Get devices associated with the tenant
            netbox_devices = self.get_devices_by_tenant(tenant.id)

            # Get all device types
            netbox_device_types = self.get_device_types()

            if dry_run or verbose:
                print_section("Netbox status", color="blue")
                print(f"Tenant='{tenant_name}'")
                print(f"ID['{tenant_name}']: {tenant.id}")
                print(f"Site['{tenant_name}']: {site.name}")
                print(
                    f"Devices['{tenant_name}']: {[device.name for device in netbox_devices]}"
                )
                print(
                    f"Device_types['{tenant_name}']: {[device_type.model for device_type in netbox_device_types]}"
                )

            return tenant, site, netbox_devices, netbox_device_types

        except Exception as e:
            raise RuntimeError(f"Failed to fetch NetBox data: {e}")


if __name__ == "__main__":
    NETBOX_URL = os.getenv("netbox_url")
    API_TOKEN = os.getenv("netbox_api_token")

    # Initialize the NetBoxAPI class with dependency injection
    netbox_api = NetBoxAPI(NETBOX_URL, API_TOKEN)

    # Run the test function
    netbox_api.test(tenant_name="Konstfack")
